##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for creating and defining Quil programs.
"""

from .quilbase import (InstructionGroup,
                       Addr,
                       While,
                       If,
                       DefGate,
                       Gate,
                       Measurement,
                       merge_resource_managers)

from .gates import MEASURE, STANDARD_GATES
from math import pi


class Program(InstructionGroup):
    def __init__(self, *instructions):
        super(Program, self).__init__()
        self.inst(*instructions)
        self.defined_gates = []

    def synthesize(self, resource_manager=None):
        self.resource_manager.reset()
        return super(Program, self).synthesize(resource_manager)

    def __add__(self, other):
        """
        Concatenate two programs together, returning a new one.

        :param Program other: Another program or instruction to concatenate to this one.
        :return: A newly concatenated program.
        :rtype: Program
        """
        if isinstance(other, Program):
            p = Program()
            p.defined_gates = self.defined_gates + other.defined_gates
            p.actions = self.actions + other.actions
            p.resource_manager = merge_resource_managers(self.resource_manager,
                                                         other.resource_manager)
            return p
        else:
            return super(Program, self).__add__(other)

    def __getitem__(self, index):
        """
        Allows indexing into the program to get an action.
        :param index: The action at the specified index.
        :return:
        """
        return self.actions[index]

    def __iter__(self):
        """
        Allow built in iteration through a program's actions, e.g. [a for a in Program(X(0))]
        :return:
        """
        return self.actions.__iter__()

    def get_qubits(self):
        """
        :return: a set of all the qubit indices allocated in this program, synthesizing freely
        allocated qubits if neccessary.
        :rtype: set
        """
        allocated_qubits = set()
        self.synthesize()
        for ii, action in self:
            if isinstance(action, Gate):
                qubit_indices = {qq.index() for qq in action.arguments}
            elif isinstance(action, Measurement):
                qubit_indices = {action.arguments[0].index()}
            else:
                continue
            allocated_qubits = set.union(allocated_qubits, qubit_indices)
        return allocated_qubits

    def defgate(self, name, matrix):
        """
        Define a new static gate.

        :param string name: The name of the gate.
        :param array-like matrix: List of lists or Numpy 2d array.
        :return: The Program instance.
        :rtype: Program
        """
        self.defined_gates.append(DefGate(name, matrix))
        return self

    def measure(self, qubit_index, classical_reg):
        """
        Measures a qubit at qubit_index and puts the result in classical_reg

        :param int qubit_index: The address of the qubit to measure.
        :param int classical_reg: The address of the classical bit to store the result.
        :returns: The Quil Program with the appropriate measure instruction appended, e.g.
                  MEASURE 0 [1]
        :rtype: Program
        """
        return self.inst(MEASURE(qubit_index, classical_reg))

    def measure_all(self, *qubit_reg_pairs):
        """
        Measures many qubits into their specified classical bits, in the order
        they were entered.

        :param Tuple qubit_reg_pairs: Tuples of qubit indices paired with classical bits.
        :return: The Quil Program with the appropriate measure instructions appended, e.g.
                  MEASURE 0 [1]
                  MEASURE 1 [2]
                  MEASURE 2 [3]
        :rtype: Program
        """
        for qubit_index, classical_reg in qubit_reg_pairs:
            self.inst(MEASURE(qubit_index, classical_reg))
        return self

    def while_do(self, classical_reg, q_program):
        """
        While a classical register at index classical_reg is 1, loop q_program

        :param int classical_reg: The classical register to check
        :param Program q_program: The Quil program to loop.
        :return: The Quil Program with the loop instructions added.
        :rtype: Program
        """
        w_loop = While(Addr(classical_reg))
        w_loop.Body.inst(q_program)
        return self.inst(w_loop)

    def if_then(self, classical_reg, if_program, else_program=None):
        """
        If the classical register at index classical reg is 1, run if_program, else run
        else_program.

        :param int classical_reg: The classical register to check as the condition
        :param Program if_program: A Quil program to execute if classical_reg is 1
        :param Program else_program: A Quil program to execute if classical_reg is 0. This
            argument is optional and defaults to an empty Program.
        :returns: The Quil Program with the branching instructions added.
        :rtype: Program
        """

        else_program = else_program if else_program is not None else Program()

        branch = If(Addr(classical_reg))
        branch.Then.inst(if_program)
        branch.Else.inst(else_program)
        return self.inst(branch)

    def dagger(self, inv_dict=None, suffix="-INV"):
        """
        Creates the conjugate transpose of the Quil program. The program must not
        contain any irreversible actions (measurement, control flow, qubit allocation).

        :return: The Quil program's inverse
        :rtype: Program

        """

        for action in self.actions:
            assert action[0] == 0, "Program must be valid Protoquil"
            gate = action[1]
            assert not isinstance(gate, Measurement), "Program cannot contain measurements"
            assert not isinstance(gate, While) and not isinstance(gate, If), \
                "Program cannot contain control flow"

        daggered = Program()

        for gate in self.defined_gates:
            if inv_dict is None or gate.name not in inv_dict:
                daggered.defgate(gate.name + suffix, gate.matrix.T.conj())

        for action in self.actions[::-1]:
            gate = action[1]
            if gate.operator_name in STANDARD_GATES:
                if gate.operator_name == "S":
                    daggered.inst(STANDARD_GATES["PHASE"](-pi / 2, *gate.arguments))
                elif gate.operator_name == "T":
                    daggered.inst(STANDARD_GATES["RZ"](pi / 4, *gate.arguments))
                elif gate.operator_name == "ISWAP":
                    daggered.inst(STANDARD_GATES["PSWAP"](pi / 2, *gate.arguments))
                else:
                    negated_params = list(map(lambda x: -1 * x, gate.parameters))
                    daggered.inst(STANDARD_GATES[gate.operator_name](*(negated_params + gate.arguments)))
            else:
                if inv_dict is None or gate.operator_name not in inv_dict:
                    gate_inv_name = gate.operator_name + suffix
                else:
                    gate_inv_name = inv_dict[gate.operator_name]

                daggered.inst(tuple([gate_inv_name] + gate.arguments))

        return daggered

    def out(self):
        """
        Converts the Quil program to a readable string.

        :return: String form of a program
        :rtype: string
        """
        s = ""
        for dg in self.defined_gates:
            s += dg.out()
            s += "\n"
        s += super(Program, self).out()
        return s


def merge_programs(prog_list):
    """
    Merges a list of pyQuil programs into a single one by appending them in sequence

    :param list prog_list: A list of pyquil programs
    :return: a single pyQuil program
    :rtype: Program
    """
    return sum(prog_list, Program())



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Contains the core pyQuil objects that correspond to Quil instructions.
"""

import numpy as np
from copy import deepcopy
from .quil_atom import QuilAtom
from .slot import Slot
from .resource_manager import (AbstractQubit, DirectQubit, Qubit,
                               ResourceManager, check_live_qubit, merge_resource_managers)
from six import integer_types

allow_raw_instructions = True
"""
Allow constructing programs containing raw instructions.
"""


def issubinstance(x, cls):
    """
    Checks if class x is an instance or subclass of cls.
    """
    return isinstance(x, cls) or issubclass(x.__class__, cls)


# These are the first values to a 2-tuple.
# This indicates all regular Quil instructions except resource management.
ACTION_INSTALL_INSTRUCTION = 0
# These are for resource management.
ACTION_INSTANTIATE_QUBIT = 1
ACTION_RELEASE_QUBIT = 2


def action(type, obj):
    return (type, obj)


def action_install(obj):
    return action(ACTION_INSTALL_INSTRUCTION, obj)


def format_matrix_element(element):
    """
    Formats a parameterized matrix element.

    :param element: {int, float, complex, str} The parameterized element to format.
    """
    if isinstance(element, integer_types) or isinstance(element, (float, complex)):
        return format_parameter(element)
    elif isinstance(element, str):
        return element
    else:
        assert False, "Invalid matrix element: %r" % element


def format_parameter(element):
    """
    Formats a particular parameter.

    :param element: {int, float, long, complex, Slot} Formats a parameter for Quil output.
    """
    if isinstance(element, integer_types) or isinstance(element, float):
        return repr(element)
    elif isinstance(element, complex):
        r = element.real
        i = element.imag
        if i < 0:
            return repr(r) + "-" + repr(abs(i)) + "i"
        else:
            return repr(r) + "+" + repr(i) + "i"
    elif isinstance(element, Slot):
        return format_parameter(element.value())
    assert False, "Invalid parameter: %r" % element


def unpack_qubit(qubit):
    """
    Get a qubit from an object.

    :param qubit: An int or AbstractQubit.
    :return: An AbstractQubit instance
    """
    if isinstance(qubit, integer_types):
        return DirectQubit(qubit)
    elif isinstance(qubit, AbstractQubit):
        return qubit
    else:
        raise TypeError("qubit should be an int or AbstractQubit instance")


class Addr(QuilAtom):
    """
    Representation of a classical bit address.

    :param int value: The classical address.
    """

    def __init__(self, value):
        if not isinstance(value, integer_types) or value < 0:
            raise TypeError("Addr value must be a non-negative int")
        self.address = value

    def __repr__(self):
        return "<Addr {0}>".format(self.address)

    def __str__(self):
        return "[{0}]".format(self.address)


class Label(QuilAtom):
    """
    Representation of a label.

    :param string label_name: The label name.
    """

    def __init__(self, label_name):
        self.name = label_name

    def __repr__(self):
        return "<Label {0}>".format(repr(self.name))

    def __str__(self):
        return "@" + self.name


class QuilAction(object):
    """
    Representation of some executable code, i.e., something that can be
    synthesized into final Quil instructions.
    """

    def synthesize(self, resource_manager=None):
        raise NotImplementedError()


class AbstractInstruction(QuilAction):
    """
    Abstract class for representing single instructionos.
    """

    def synthesize(self, resource_manager=None):
        return [self]

    def out(self):
        return NotImplementedError()


class DefGate(AbstractInstruction):
    """
    A DEFGATE directive.

    :param string name: The name of the newly defined gate.
    :param array-like matrix: {list, nparray, np.matrix} The matrix defining this gate.
    """

    def __init__(self, name, matrix):
        assert isinstance(name, str)
        assert isinstance(matrix, (list, np.ndarray, np.matrix))
        if isinstance(matrix, list):
            rows = len(matrix)
            assert all([len(row) == rows for row in matrix]), "Matrix must be square."
        elif isinstance(matrix, (np.ndarray, np.matrix)):
            rows, cols = matrix.shape
            assert rows == cols, "Matrix must be square."
        else:
            raise TypeError("Matrix argument must be a list or NumPy array/matrix")

        if 0 != rows & (rows - 1):
            raise AssertionError("Dimension of matrix must be a power of 2, got {0}"
                                 .format(rows))
        self.name = name
        self.matrix = np.asarray(matrix)

        is_unitary = np.allclose(np.eye(rows), self.matrix.dot(self.matrix.T.conj()))
        if not is_unitary:
            raise AssertionError("Matrix must be unitary.")

    def out(self):
        """
        Prints a readable Quil string representation of this gate.

        :returns: String representation of a gate
        :rtype: string
        """
        result = "DEFGATE %s:\n" % (self.name)
        for row in self.matrix:
            result += "    "
            fcols = [format_matrix_element(col) for col in row]
            result += ", ".join(fcols)
            result += "\n"
        return result

    def get_constructor(self):
        """
        :returns: A function that constructs this gate on variable qubit indices. E.g.
                  `mygate.get_constructor()(1) applies the gate to qubit 1.`
        """
        return lambda *qubits: Gate(name=self.name, params=[], qubits=list(map(unpack_qubit, qubits)))

    def num_args(self):
        """
        :return: The number of qubit arguments the gate takes.
        :rtype: int
        """
        rows = len(self.matrix)
        return int(np.log2(rows))


class InstructionGroup(QuilAction):
    """
    Representation of a sequence of instructions that can be synthesized into a Quil program.
    """

    def __init__(self, resource_manager=None):
        self.actions = []
        if resource_manager is None:
            self.resource_manager = ResourceManager()
        else:
            self.resource_manager = resource_manager

    def synthesize(self, resource_manager=None):
        synthesized = []
        for action_type, obj in self.actions:
            if action_type == ACTION_INSTALL_INSTRUCTION:
                synthesized.extend(obj.synthesize(self.resource_manager))
            elif action_type == ACTION_INSTANTIATE_QUBIT:
                self.resource_manager.instantiate(obj)
            elif action_type == ACTION_RELEASE_QUBIT:
                self.resource_manager.uninstantiate_index(obj.assignment)
            else:
                raise RuntimeError("encountered invalid action")

        return synthesized

    def __str__(self):
        return self.out()

    def out(self):
        instrs = self.synthesize()
        s = ""
        for instr in instrs:
            s += instr.out() + "\n"
        return s

    def alloc(self):
        """
        Get a new qubit.

        :return: A qubit.
        :rtype: Qubit
        """
        qubit = self.resource_manager.allocate_qubit()
        self.actions.append(action(ACTION_INSTANTIATE_QUBIT, qubit))

        return qubit

    def free(self, qubit):
        """
        Free a qubit.

        :param AbstractQubit q: An AbstractQubit instance.
        """
        check_live_qubit(qubit)

        if qubit.resource_manager != self.resource_manager:
            raise RuntimeError("qubit is managed by a different instruction group")

        self.actions.append(action(ACTION_RELEASE_QUBIT, qubit))
        self.resource_manager.free_qubit(qubit)

    def inst(self, *instructions):
        """
        Mutates the Program object by appending new instructions.

        :param instructions: A list of Instruction objects, e.g. Gates
        :return: self
        """
        for instruction in instructions:
            if isinstance(instruction, list):
                self.inst(*instruction)
            elif isinstance(instruction, tuple):
                if len(instruction) == 0:
                    raise ValueError("tuple should have at least one element")
                elif len(instruction) == 1:
                    self.actions.append(action_install(Instr(instruction[0], [], [])))
                else:
                    op = instruction[0]
                    params = []
                    possible_params = instruction[1]
                    rest = instruction[2:]
                    if isinstance(possible_params, list):
                        params = possible_params
                    else:
                        rest = [possible_params] + list(rest)
                    self.actions.append(action_install(Instr(op, params, rest)))
            elif isinstance(instruction, str):
                self.actions.append(action_install(RawInstr(instruction)))
            elif issubinstance(instruction, QuilAction):
                self.actions.append(action_install(instruction))
            elif issubinstance(instruction, InstructionGroup):
                self.resource_manager = merge_resource_managers(self.resource_manager,
                                                                instruction.resource_manager)
                self.actions.extend(list(instruction.actions))
            else:
                raise TypeError("Invalid instruction: {}".format(instruction))
        # Return self for method chaining.
        return self

    def __add__(self, instruction):
        p = deepcopy(self)
        return p.inst(instruction)

    def pop(self):
        """
        Pops off the last instruction.

        :return: The (action, instruction) pair for the instruction that was popped.
        :rtype: tuple
        """
        if 0 != len(self.actions):
            return self.actions.pop()

    def extract_qubits(self):
        """
        Return all qubit addresses involved in the instruction group.
        :return: Set of qubits.
        :rtype: set
        """
        qubits = set()
        for jj, act_jj in self.actions:
            if jj == ACTION_INSTALL_INSTRUCTION:
                if isinstance(act_jj, Instr):
                    qubits = qubits | act_jj.qubits()
                elif isinstance(act_jj, If):
                    qubits = qubits | act_jj.Then.extract_qubits() | act_jj.Else.extract_qubits()
                elif isinstance(act_jj, While):
                    qubits = qubits | act_jj.Body.extract_qubits()
                elif isinstance(act_jj, InstructionGroup):
                    qubits = qubits | act_jj.extract_qubits()
                elif isinstance(act_jj, (JumpTarget, JumpConditional, SimpleInstruction,
                                         UnaryClassicalInstruction, BinaryClassicalInstruction,
                                         Jump)):
                    continue
                else:
                    raise ValueError(type(act_jj))
            elif jj in (ACTION_INSTANTIATE_QUBIT, ACTION_RELEASE_QUBIT):
                continue
        return qubits


class JumpTarget(AbstractInstruction):
    """
    Representation of a target that can be jumped to.
    """

    def __init__(self, label):
        if not isinstance(label, Label):
            raise TypeError("label must be a Label")
        self.label = label

    def __repr__(self):
        return "<JumpTarget {0}>".format(str(self.label))

    def out(self):
        return "LABEL {0}".format(str(self.label))


class JumpConditional(AbstractInstruction):
    """
    Abstract representation of an conditional jump instruction.
    """

    def __init__(self, target, condition):
        if not isinstance(target, Label):
            raise TypeError("target should be a Label")
        if not isinstance(condition, Addr):
            raise TypeError("condition should be an Addr")
        self.target = target
        self.condition = condition

    def __str__(self):
        return self.out()

    def out(self):
        return "%s %s %s" % (self.op, self.target, self.condition)


class JumpWhen(JumpConditional):
    """
    The JUMP-WHEN instruction.
    """
    op = "JUMP-WHEN"


class JumpUnless(JumpConditional):
    """
    The JUMP-UNLESS instruction.
    """
    op = "JUMP-UNLESS"


class SimpleInstruction(AbstractInstruction):
    """
    Abstract class for simple instructions with no arguments.
    """

    def __str__(self):
        return self.out()

    def out(self):
        return self.op


class Halt(SimpleInstruction):
    """
    The HALT instruction.
    """
    op = "HALT"


class Wait(SimpleInstruction):
    """
    The WAIT instruction.
    """
    op = "WAIT"


class Reset(SimpleInstruction):
    """
    The RESET instruction.
    """
    op = "RESET"


class Nop(SimpleInstruction):
    """
    The RESET instruction.
    """
    op = "NOP"


class UnaryClassicalInstruction(AbstractInstruction):
    """
    The abstract class for unary classical instructions.
    """

    def __init__(self, target):
        if not isinstance(target, Addr):
            raise TypeError("target operand should be an Addr")
        self.target = target

    def __str__(self):
        return self.out()

    def out(self):
        return "%s %s" % (self.op, self.target)


class ClassicalTrue(UnaryClassicalInstruction):
    op = "TRUE"


class ClassicalFalse(UnaryClassicalInstruction):
    op = "FALSE"


class ClassicalNot(UnaryClassicalInstruction):
    op = "NOT"


class BinaryClassicalInstruction(AbstractInstruction):
    """
    The abstract class for binary classical instructions.
    """

    def __init__(self, left, right):
        if not isinstance(left, Addr):
            raise TypeError("left operand should be an Addr")
        if not isinstance(right, Addr):
            raise TypeError("right operand should be an Addr")
        self.left = left
        self.right = right

    def __str__(self):
        return self.out()

    def out(self):
        return "%s %s %s" % (self.op, self.left, self.right)


class ClassicalAnd(BinaryClassicalInstruction):
    op = "AND"


class ClassicalOr(BinaryClassicalInstruction):
    op = "OR"


class ClassicalMove(BinaryClassicalInstruction):
    op = "MOVE"


class ClassicalExchange(BinaryClassicalInstruction):
    op = "EXCHANGE"


class Jump(AbstractInstruction):
    """
    Representation of an unconditional jump instruction (JUMP).
    """

    def __init__(self, target):
        if not isinstance(target, Label):
            raise TypeError("target should be a Label")
        self.target = target

    def __str__(self):
        return self.out()

    def out(self):
        return "JUMP %s" % self.target


label_counter = 0


def reset_label_counter():
    global label_counter
    label_counter = 0


def gen_label(prefix="L"):
    """
    Generate a fresh label.

    :param string prefix: An optional prefix for the label name.
    :return: A new Label instance.
    :rtype: Label
    """
    global label_counter
    label_counter += 1
    return Label(prefix + str(label_counter))


class RawInstr(AbstractInstruction):
    """
    A raw instruction represented as a string.
    """

    def __init__(self, instr_str):
        if not isinstance(instr_str, str):
            raise TypeError("Raw instructions require a string.")
        if not allow_raw_instructions:
            raise RuntimeError("Raw instructions are not allowed. Consider changing"
                               "the variable `allow_raw_instructions` to `True`.")
        self.instr = instr_str

    def out(self):
        return self.instr

    def __repr__(self):
        return '<RawInstr>'

    def __str__(self):
        return self.instr


class Instr(AbstractInstruction):
    """
    Representation of an instruction represented by an operator, parameters, and arguments.
    """

    def __init__(self, op, params, args):
        if not isinstance(op, str):
            raise TypeError("op must be a string")
        self.operator_name = op
        self.parameters = params
        self.arguments = []
        if 0 != len(args):
            if isinstance(args[0], list):
                self.parameters = None if 0 == len(args[0]) else args[0]
                self.arguments = args[1:]
            else:
                self.arguments = args

    def __str__(self):
        return "<Instr {0}>".format(self.operator_name)

    def __eq__(self, other):
        return self.out() == other.out()

    def __ne__(self, other):
        # x!=y and x<>y call __ne__() instead of negating __eq__
        # This is only a weirdness in python2 as in python 3 __ne__ defaults to the inversion of
        # __eq__
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.out())

    def synthesize(self, resource_manager=None):
        if resource_manager is not None:
            self.make_qubits_known(resource_manager)
        return [self]

    def out(self):
        def format_params(params):
            if not params:
                return ""
            else:
                return "(" + ",".join(map(format_parameter, params)) + ")"

        def format_args(args):
            if 0 == len(args):
                return ""
            else:
                return " " + " ".join([str(arg) for arg in args])

        if self.parameters:
            return self.operator_name + format_params(self.parameters) + format_args(self.arguments)
        else:
            return self.operator_name + format_args(self.arguments)

    def make_qubits_known(self, rm):
        """
        Make the qubits involved with this instruction known to a ResourceManager.

        :param ResourceManager rm: A ResourceManager object.
        """
        if not isinstance(rm, ResourceManager):
            raise TypeError("rm should be a ResourceManager")

        for arg in self.arguments:
            if isinstance(arg, DirectQubit):
                current = rm.in_use.get(arg.index(), False)
                rm.in_use[arg.index()] = arg
                # re-instantiate the qubit
                if current and isinstance(current, Qubit):
                    rm.instantiate(current)

    def qubits(self):
        """
        The qubits this instruction affects.

        :return: Set of qubit indexes.
        :rtype: set
        """
        qubits = set()
        for arg in self.arguments:
            if issubinstance(arg, AbstractQubit):
                qubits.add(arg.index())
        return qubits


class Gate(Instr):
    """
    This is the pyQuil object for a quantum gate instruction.
    """

    def __init__(self, name, params, qubits):
        for qubit in qubits:
            check_live_qubit(qubit)
        super(Gate, self).__init__(name, params, qubits)

    def __repr__(self):
        return "<Gate: " + self.out() + ">"

    def __str__(self):
        return self.out()


class Measurement(Instr):
    """
    This is the pyQuil object for a Quil measurement instruction.
    """

    def __init__(self, qubit, classical_reg=None):
        check_live_qubit(qubit)

        if classical_reg is None:
            args = (qubit,)
        elif isinstance(classical_reg, Addr):
            args = (qubit, classical_reg)
        else:
            raise TypeError("classical_reg should be None or an Addr instance")

        super(Measurement, self).__init__("MEASURE",
                                          params=None,
                                          args=args)

        self.classical_reg = classical_reg


class While(QuilAction):
    """
    Representation of a ``while`` construct. To use, initialize with an
    address to branch on, and use ``self.Body.inst()`` to add instructions to
    the body of the loop.
    """

    def __init__(self, condition):
        if not isinstance(condition, Addr):
            raise TypeError("condition must be an Addr")
        super(While, self).__init__()
        self.condition = condition
        self.Body = InstructionGroup()

    def synthesize(self, resource_manager=None):
        # WHILE [c]:
        #    instr...
        #
        # =>
        #
        #   LABEL @START
        #   JUMP-UNLESS @END [c]
        #   instr...
        #   JUMP @START
        #   LABEL @END
        label_start = gen_label("START")
        label_end = gen_label("END")
        insts = list()
        insts.append(JumpTarget(label_start))
        insts.append(JumpUnless(target=label_end, condition=self.condition))
        insts.extend(self.Body.synthesize())
        insts.append(Jump(target=label_start))
        insts.append(JumpTarget(label_end))
        return insts


class If(QuilAction):
    """
    Representation of an ``if`` construct. To use, initialize with an address
    to be branched on, and add instructions to ``self.Then`` and ``self.Else``
    for the corresponding branches.
    """

    def __init__(self, condition):
        if not isinstance(condition, Addr):
            raise TypeError("condition must be an Addr")
        super(If, self).__init__()
        self.condition = condition
        self.Then = InstructionGroup()
        self.Else = InstructionGroup()

    def synthesize(self, resource_manager=None):
        # IF [c]:
        #    instrA...
        # ELSE:
        #    instrB...
        #
        # =>
        #
        #   JUMP-WHEN @THEN [c]
        #   instrB...
        #   JUMP @END
        #   LABEL @THEN
        #   instrA...
        #   LABEL @END
        label_then = gen_label("THEN")
        label_end = gen_label("END")
        insts = list()
        insts.append(JumpWhen(target=label_then, condition=self.condition))
        insts.extend(self.Else.synthesize())
        insts.append(Jump(target=label_end))
        insts.append(JumpTarget(label_then))
        insts.extend(self.Then.synthesize())
        insts.append(JumpTarget(label_end))
        return insts



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for facilitating connections to the QVM / QPU.
"""

from __future__ import print_function
from requests.adapters import HTTPAdapter
from six import integer_types
from urllib3.util.retry import Retry
import json
import requests
import sys

from pyquil.config import PyquilConfig
from pyquil.job_results import JobResult, WavefunctionResult, recover_complexes
from pyquil.quil import Program


def add_noise_to_payload(payload, gate_noise, measurement_noise):
    """
    Set the gate noise and measurement noise of a payload.

    :param dict payload: Dictionary of run information.
    :param gate_noise: Probability of a noise gate being applied at each time step.
    :param measurement_noise: Probability of a noise gate being applied before measurement.
    """
    if gate_noise is not None:
        payload["gate-noise"] = gate_noise
    if measurement_noise is not None:
        payload["measurement-noise"] = measurement_noise


def add_rng_seed_to_payload(payload, seed):
    """
    Add a random seed to the payload.

    :param payload: JSON payload.
    :param int seed: A non-negative integer.
    """
    if seed is not None:
        payload['rng-seed'] = seed


def _validate_noise_probabilities(noise_parameter):
    """
    Is noise_parameter a valid specification of noise probabilities for depolarizing noise?

    :param list noise_parameter: List of noise parameter values to be validated.
    """
    if not noise_parameter:
        return
    if not isinstance(noise_parameter, list):
        raise TypeError("noise_parameter must be a list")
    if any([not isinstance(value, float) for value in noise_parameter]):
        raise TypeError("noise_parameter values should all be floats")
    if len(noise_parameter) != 3:
        raise ValueError("noise_parameter lists must be of length 3")
    if sum(noise_parameter) > 1 or sum(noise_parameter) < 0:
        raise ValueError("sum of entries in noise_parameter must be between 0 and 1 (inclusive)")
    if any([value < 0 for value in noise_parameter]):
        raise ValueError("noise_parameter values should all be non-negative")


def validate_run_items(run_items):
    """
    Check the validity of classical addresses / qubits for the payload.

    :param list run_items: List of classical addresses or qubits to be validated.
    """
    if not isinstance(run_items, list):
        raise TypeError("run_items must be a list")
    if any([not isinstance(i, integer_types) for i in run_items]):
        raise TypeError("run_items list must contain integer values")


TYPE_EXPECTATION = "expectation"
TYPE_MULTISHOT = "multishot"
TYPE_MULTISHOT_MEASURE = "multishot-measure"
TYPE_WAVEFUNCTION = "wavefunction"


class JobConnection(object):
    """
    Represents a connection to the QVM.
    """

    #################################################################################
    # PS: WELCOME TO THE SOURCE! Do you want fast-tracked access to a Rigetti API key?
    #    Email us at batman@rigetti.com and tell us what you'd like to do.  We're looking
    #    for people like you to join our team: https://jobs.lever.co/rigetti
    #################################################################################

    def __init__(self, endpoint=None, api_key=None, user_id=None, gate_noise=None,
                 measurement_noise=None, num_retries=3, random_seed=None):
        """
        Constructor for JobConnection. Sets up any necessary security, and establishes the noise
        model to use.

        :param endpoint: The endpoint of the server (default behavior is to read from config file)
        :param api_key: The key to the Forest API Gateway (default behavior is to read from config file)
        :param user_id: Your userid for Forest (default behavior is to read from config file)
        :param gate_noise: A list of three numbers [Px, Py, Pz] indicating the probability of an X,
                           Y, or Z gate getting applied to each qubit after a gate application or
                           reset. (default None)
        :param measurement_noise: A list of three numbers [Px, Py, Pz] indicating the probability of
                                  an X, Y, or Z gate getting applied before a a measurement.
                                  (default None)
        :param num_retries: The number of times to retry a request when faced with an HTTP response
                            code in [502, 503, 504, 521, 523] (default 3)
        :param random_seed: A seed for the QVM's random number generators. Either None (for an
                            automatically generated seed) or a non-negative integer.
        """
        # Once these are set, they should not ever be cleared/changed/touched.
        # Make a new JobConnection() if you need that.
        _validate_noise_probabilities(gate_noise)
        _validate_noise_probabilities(measurement_noise)
        self.gate_noise = gate_noise
        self.measurement_noise = measurement_noise

        self.session = requests.Session()

        # We need this to get binary payload for the wavefunction call.
        self.session.headers.update({"Accept": "application/octet-stream"})

        self.num_retries = num_retries
        retry_adapter = HTTPAdapter(max_retries=Retry(total=num_retries,
                                                      method_whitelist=['POST'],
                                                      status_forcelist=[502, 503, 504, 521, 523],
                                                      backoff_factor=0.2,
                                                      raise_on_status=False))
        self.session.mount("http://", retry_adapter)
        self.session.mount("https://", retry_adapter)

        config = PyquilConfig()
        self.endpoint = endpoint if endpoint else config.endpoint
        self.api_key = api_key if api_key else config.api_key
        self.user_id = user_id if user_id else config.user_id
        self.headers = {
            'Content-Type': 'application/json; charset=utf-8',
            'x-api-key': self.api_key,
            'x-user-id': self.user_id,
        }

        if random_seed is None:
            self.random_seed = None
        elif isinstance(random_seed, integer_types) and random_seed >= 0:
            self.random_seed = random_seed
        else:
            raise TypeError("random_seed should be None or a non-negative int or long.")

        self.machine = 'QVM'  # default to a QVM Machine Connection

    def post_json(self, jd, headers=None, route=""):
        """
        Post JSON to the Forest endpoint.

        :param dict jd: JSON.
        :param dict headers: The headers for the post request.
        :param str route: The route to append to the endpoint, e.g. "/job"
        :return: A non-error response.
        """
        if headers is None:
            headers = self.headers
        url = self.endpoint + route
        res = self.session.post(url, json=jd, headers=headers)

        # Print some nice info for unauthorized/permission errors.
        if res.status_code == 401 or res.status_code == 403:
            print("! ERROR:\n"
                  "!   There was an issue validating your forest account.\n"
                  "!   Have you run the pyquil-config-setup command yet?\n"
                  "! The server came back with the following information:\n"
                  "%s\n%s\n%s" % ("=" * 80, res.text, "=" * 80),
                  file=sys.stderr)
            print("! If you suspect this to be a bug in pyQuil or Rigetti Forest,\n"
                  "! then please describe the problem in a GitHub issue at:\n!\n"
                  "!      https://github.com/rigetticomputing/pyquil/issues\n",
                  file=sys.stderr)

        # Print some nice info for invalid input or internal server errors.
        if res.status_code == 400 or res.status_code >= 500:
            print("! ERROR:\n"
                  "!   Server caught an error. This could be due to a bug in the server\n"
                  "!   or a bug in your code. The server came back with the following\n"
                  "!   information:\n"
                  "%s\n%s\n%s" % ("=" * 80, res.text, "=" * 80),
                  file=sys.stderr)
            print("! If you suspect this to be a bug in pyQuil or Rigetti Forest,\n"
                  "! then please describe the problem in a GitHub issue at:\n!\n"
                  "!      https://github.com/rigetticomputing/pyquil/issues\n",
                  file=sys.stderr)

        res.raise_for_status()
        return res

    def wrap_payload_into_message(self, payload):
        """
        Wraps the payload into a Forest query.

        :param dict payload:
        :return: A JSON dictionary to post as a Forest query.
        """
        message = {
            'machine': self.machine,
            'program': payload,
            'userId': self.user_id,
        }
        return message

    def post_job(self, program, headers=None):
        """
        Post a Job to the Forest endpoint.

        :param program:
        :param headers:
        :return: A non-error response.
        """
        message = self.wrap_payload_into_message(program)
        return self.post_json(message, headers, route="/job")

    def get_job(self, job_result):
        """
        :param JobResult job_result:
        :return: fills in the result in the JobResult object
        """
        url = self.endpoint + ("/job/%s" % (job_result.job_id()))
        res = requests.get(url, headers=self.headers)
        result = res.json()
        return job_result._update(res.ok, result)

    def ping(self):
        """
        Ping Forest.

        :return: Should get "ok" back.
        :rtype: string
        """
        res = self.session.get(self.endpoint + "/check")
        return str(json.loads(res.text)["rc"])

    def version(self):
        """
        Query the QVM version.

        :return: None
        """
        raise DeprecationWarning("Version checks have been deprecated.")

    def process_response(self, res):
        """
        :param res: A response object from a request
        :return: A JobResult filled in with the response data
        :rtype: JobResult
        """
        return JobResult.load_res(self, res)

    def process_wavefunction_response(self, res, payload):
        """
        Wavefunctions are processed differently as they are byte encoded.
        :param res: A response object from a request
        :param payload: The payload that was used to make that request
        :return: A WavefunctionResult with the response data filled in
        :rtype: WavefunctionResult
        """
        result = res.json()
        return WavefunctionResult(self, res.ok, result=result, payload=payload)

    def wavefunction(self, quil_program, classical_addresses=None):
        """
        Simulate a Quil program and get the wavefunction back.

        :param Program quil_program: A Quil program.
        :param list classical_addresses: An optional list of classical addresses.
        :return: A tuple whose first element is a Wavefunction object,
                 and whose second element is the list of classical bits corresponding
                 to the classical addresses.
        :rtype: tuple
        """
        if classical_addresses is None:
            classical_addresses = []

        if not isinstance(quil_program, Program):
            raise TypeError("quil_program must be a Quil program object")
        validate_run_items(classical_addresses)

        payload = {'type': TYPE_WAVEFUNCTION,
                   'quil-instructions': quil_program.out(),
                   'addresses': classical_addresses}
        add_noise_to_payload(payload, self.gate_noise, self.measurement_noise)
        add_rng_seed_to_payload(payload, self.random_seed)

        res = self.post_job(payload, headers=self.headers)
        return self.process_wavefunction_response(res, payload)

    def expectation(self, prep_prog, operator_programs=None):
        """
        Calculate the expectation value of operators given a state prepared by
        prep_program.

        :param Program prep_prog: Quil program for state preparation.
        :param list operator_programs: A list of PauliTerms. Default is Identity operator.
        :returns: Expectation value of the operators.
        :rtype: float
        """
        if operator_programs is None:
            operator_programs = [Program()]

        if not isinstance(prep_prog, Program):
            raise TypeError("prep_prog variable must be a Quil program object")

        payload = {'type': TYPE_EXPECTATION,
                   'state-preparation': prep_prog.out(),
                   'operators': [x.out() for x in operator_programs]}

        add_rng_seed_to_payload(payload, self.random_seed)

        res = self.post_job(payload, headers=self.headers)
        return self.process_response(res)

    def bit_string_probabilities(self, quil_program):
        """
        Simulate a Quil program and get outcome probabilities back.

        :param Program quil_program: A Quil program.
        :return: A dictionary with outcomes as keys and probabilities as values.
        :rtype: dict
        """
        return DeprecationWarning, "This function is deprecated."

    def run(self, quil_program, classical_addresses, trials=1):
        """
        Run a Quil program multiple times, accumulating the values deposited in
        a list of classical addresses.

        :param Program quil_program: A Quil program.
        :param list classical_addresses: A list of addresses.
        :param int trials: Number of shots to collect.
        :return: A list of lists of bits. Each sublist corresponds to the values
                 in `classical_addresses`.
        :rtype: list
        """
        if not isinstance(quil_program, Program):
            raise TypeError("quil_program must be a Quil program object")
        validate_run_items(classical_addresses)
        if not isinstance(trials, integer_types):
            raise TypeError("trials must be an integer")

        payload = {"type": TYPE_MULTISHOT,
                   "addresses": classical_addresses,
                   "trials": trials,
                   "quil-instructions": quil_program.out()}

        add_noise_to_payload(payload, self.gate_noise, self.measurement_noise)
        add_rng_seed_to_payload(payload, self.random_seed)

        res = self.post_job(payload, headers=self.headers)
        return self.process_response(res)

    def run_and_measure(self, quil_program, qubits, trials=1):
        """
        Run a Quil program once to determine the final wavefunction, and measure multiple times.

        :param Program quil_program: A Quil program.
        :param list qubits: A list of qubits.
        :param int trials: Number of shots to collect.
        :return: A list of a list of bits.
        :rtype: list
        """
        if not isinstance(quil_program, Program):
            raise TypeError("quil_program must be a Quil program object")
        validate_run_items(qubits)
        if not isinstance(trials, integer_types):
            raise TypeError("trials must be an integer")

        payload = {"type": TYPE_MULTISHOT_MEASURE,
                   "qubits": qubits,
                   "trials": trials,
                   "quil-instructions": quil_program.out()}

        add_noise_to_payload(payload, self.gate_noise, self.measurement_noise)
        add_rng_seed_to_payload(payload, self.random_seed)

        res = self.post_job(payload, headers=self.headers)
        return self.process_response(res)


class SyncConnection(JobConnection):
    """
    The SyncConnection makes a synchronous connection to the Forest API.
    """

    def ping(self):
        """
        Ping the QVM.
        :return: Should get "pong" back.
        :rtype: string
        """
        payload = {"type": "ping"}
        res = self.post_json(payload)
        return str(res.text)

    def version(self):
        """
        Query the QVM version.
        :return: The current version of the QVM.
        :rtype: string
        """
        payload = {"type": "version"}
        res = self.post_json(payload)
        return str(res.text)

    def post_job(self, program, headers=None):
        """
        Post a synchronous Job to the QVM endpoint.

        :param program:
        :param headers:
        :return: A non-error response.
        """
        return self.post_json(program, headers)

    def process_response(self, res):
        """
        :param res: A response object from a request
        :return: The json dictionary of the response
        :rtype: dict
        """
        return json.loads(res.text)

    def process_wavefunction_response(self, res, payload):
        """
        :param res: A response object from a request
        :return: The json dictionary of the response
        :rtype: dict
        """
        return recover_complexes(res.content, classical_addresses=payload['addresses'])

    def get_job(self, job_result):
        raise NotImplementedError



"""
Module containing the Wavefunction object and methods for working with wavefunctions.
"""


class Wavefunction(object):

    def __init__(self, amplitude_vector):
        """
        Initializes a wavefunction
        :param amplitude_vector: A numpy array of complex amplitudes
        """
        if len(amplitude_vector) == 0 or len(amplitude_vector) & (len(amplitude_vector) - 1) != 0:
            raise TypeError("Amplitude vector must have a length that is a power of two")

        self.amplitudes = amplitude_vector

    def __len__(self):
        return len(self.amplitudes).bit_length() - 1

    def __iter__(self):
        return self.amplitudes.__iter__()

    def __getitem__(self, index):
        return self.amplitudes[index]

    def __setitem__(self, key, value):
        self.amplitudes[key] = value

    def __str__(self):
        return self.pretty_print(decimal_digits=10)

    def get_outcome_probs(self):
        """
        Parses a wavefunction (array of complex amplitudes) and returns a dictionary of
        outcomes and associated probabilities.

        :return: A dict with outcomes as keys and probabilities as values.
        :rtype: dict
        """
        outcome_dict = {}
        qubit_num = len(self)
        for index, amplitude in enumerate(self):
            outcome = get_bitstring_from_index(index, qubit_num)
            outcome_dict[outcome] = abs(amplitude) ** 2
        return outcome_dict

    def pretty_print_probabilities(self, decimal_digits=2):
        """
        Prints outcome probabilities, ignoring all outcomes with approximately zero probabilities
        (up to a certain number of decimal digits) and rounding the probabilities to decimal_digits.

        :param int decimal_digits: The number of digits to truncate to.
        :return: A dict with outcomes as keys and probabilities as values.
        :rtype: dict
        """
        outcome_dict = {}
        qubit_num = len(self)
        for index, amplitude in enumerate(self):
            outcome = get_bitstring_from_index(index, qubit_num)
            prob = round(abs(amplitude) ** 2, decimal_digits)
            if prob != 0.:
                outcome_dict[outcome] = prob
        return outcome_dict

    def pretty_print(self, decimal_digits=2):
        """
        Returns a string repr of the wavefunction, ignoring all outcomes with approximately zero
        amplitude (up to a certain number of decimal digits) and rounding the amplitudes to
        decimal_digits.

        :param int decimal_digits: The number of digits to truncate to.
        :return: A dict with outcomes as keys and complex amplitudes as values.
        :rtype: str
        """
        outcome_dict = {}
        qubit_num = len(self)
        pp_string = ""
        for index, amplitude in enumerate(self):
            outcome = get_bitstring_from_index(index, qubit_num)
            amplitude = round(amplitude.real, decimal_digits) + \
                round(amplitude.imag, decimal_digits) * 1.j
            if amplitude != 0.:
                outcome_dict[outcome] = amplitude
                pp_string += str(amplitude) + "|{}> + ".format(outcome)
        if len(pp_string) >= 3:
            pp_string = pp_string[:-3]  # remove the dangling + if it is there
        return pp_string

    @classmethod
    def ground(cls, qubit_num):
        """
        Constructs the groundstate wavefunction for a given number of qubits.

        :param int qubit_num:
        :return: A Wavefunction in the ground state
        :rtype: Wavefunction
        """
        container = [0] * (2**qubit_num)
        container[0] = 1.0
        return cls(container)


def get_bitstring_from_index(index, qubit_num):
    """
    Returns the bitstring in lexical order that corresponds to the given index in 0 to 2^(qubit_num)
    :param int index:
    :param int qubit_num:
    :return: the bitstring
    :rtype: str
    """
    if index > (2**qubit_num - 1):
        raise IndexError("Index {} too large for {} qubits.".format(index, qubit_num))
    return bin(index)[2:].rjust(qubit_num, '0')



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for facilitating connections to the QPU.
"""

import requests

from pyquil.api import JobConnection, TYPE_MULTISHOT, TYPE_MULTISHOT_MEASURE, validate_run_items
from pyquil.job_results import RamseyResult, RabiResult, T1Result
from pyquil.quil import Program

ENDPOINT = 'https://job.rigetti.com/beta'


class NoParametersFoundException(Exception):
    pass


class QPUConnection(JobConnection):

    def __init__(self, device_name='Device', **kwargs):
        """
        :param str device_name: Unique identifier of the device in question
        """
        super(QPUConnection, self).__init__(**kwargs)
        self.device_name = device_name
        self.machine = 'QPU'
        self.session.headers = self.headers
        self.endpoint = ENDPOINT

    def get_qubits(self):
        """
        :return: A list of active qubit ids on this device.
        :rtype: list
        """
        config_dict = self.get_info()
        device_config = config_dict[self.device_name]
        return [qq['num'] for qq in device_config['qubits']]

    def rabi(self, qubit_id):
        """
        Runs a Rabi experiment on the given qubit.

        :param int qubit_id: Unique identifier of the qubit in question
        :return: A RabiResult object
        """
        payload = self.get_rabi_params(qubit_id)
        payload.update({
            'type': 'pyquillow',
            'experiment': 'rabi',
            'qcid': qubit_id,
            'device_id': self.device_name,
        })
        res = self.post_job(payload)
        return RabiResult.load_res(self, res)

    def ramsey(self, qubit_id):
        """
        Runs a Ramsey experiment on the given qubit.

        :param int qubit_id: Unique identifier of the qubit in question
        :return: A RamseyResult object
        """
        payload = self.get_ramsey_params(qubit_id)
        payload.update({
            'type': 'pyquillow',
            'experiment': 'ramsey',
            'qcid': qubit_id,
            'device_id': self.device_name,
        })
        res = self.post_job(payload)
        return RamseyResult.load_res(self, res)

    def t1(self, qubit_id):
        """
        Runs a T1 experiment on the given qubit.

        :param int qubit_id: Unique identifier of the qubit in question
        :return: A T1Result object
        """
        payload = self.get_t1_params(qubit_id)
        payload.update({
            'type': 'pyquillow',
            'experiment': 't1',
            'qcid': qubit_id,
            'device_id': self.device_name,
        })
        res = self.post_job(payload)
        return T1Result.load_res(self, res)

    def wavefunction(self, quil_program, classical_addresses=[]):
        return NotImplementedError

    def run(self, quil_program, classical_addresses, trials=1):
        """
        Run a pyQuil program on the QPU. This functionality is in beta.

        :param Program quil_program: Quil program to run on the QPU
        :param list classical_addresses: Currently unused
        :param int trials: Number of shots to take
        :return: A job result
        :rtype: JobResult
        """
        if not isinstance(quil_program, Program):
            raise TypeError('quil_program must be a Quil program object')
        validate_run_items(classical_addresses)
        if not isinstance(trials, int):
            raise TypeError('trials must be an integer')

        payload = {'type': TYPE_MULTISHOT,
                   'addresses': classical_addresses,
                   'trials': trials,
                   'quil-instructions': quil_program.out(),
                   'device_id': self.device_name}

        res = self.post_job(payload, headers=self.headers)
        return self.process_response(res)

    def run_and_measure(self, quil_program, qubits, trials=1):
        """
        Run a pyQuil program on the QPU multiple times, measuring all the qubits in the QPU
        simultaneously at the end of the program each time. This functionality is in beta.

        :param Program quil_program: Quil program to run on the QPU
        :param list qubits: The list of qubits to return results for
        :param int trials: Number of shots to take
        :return: A job result
        :rtype: JobResult
        """
        if not isinstance(quil_program, Program):
            raise TypeError('quil_program must be a Quil program object')
        validate_run_items(qubits)
        if not isinstance(trials, int):
            raise TypeError('trials must be an integer')

        payload = {'type': TYPE_MULTISHOT_MEASURE,
                   'qubits': qubits,
                   'trials': trials,
                   'quil-instructions': quil_program.out(),
                   'device_id': self.device_name}

        res = self.post_job(payload, headers=self.headers)
        return self.process_response(res)

    def expectation(self, prep_prog, operator_programs=[Program()]):
        return NotImplementedError

    def bit_string_probabilities(self, quil_program):
        return NotImplementedError

    def get_info(self):
        """
        Gets information about what devices are currently available through the Forest API.

        :return: A JSON dictionary with configuration information.
        :rtype: dict
        """
        url = self.endpoint + '/config'
        headers = {
            'Content-Type': 'application/json; charset=utf-8',
            'x-api-key': self.api_key,
            'x-user-id': self.user_id,
        }
        res = requests.get(url, headers=headers)
        config_json = res.json()
        return config_json

    def get_params(self, qcid, func):
        """
        Get and parse the configuration information from the Forest API.

        :param int qcid: Unique identifier of the qubit in question
        :param func: A function to apply to the qubit specific JSON dictionary of config info
        :return: A dictionary with the parameter info
        :rtype: dict
        """
        config_dict = self.get_info()
        try:
            device_config = [d for d in config_dict['devices'] if d['name'] == self.device_name][0]
            device_config = device_config['qubits']
        except (KeyError, IndexError):
            raise NoParametersFoundException('Device {} not found.'.format(self.device_name))
        for qc in device_config:
            if qc['num'] == qcid:
                return func(qc)
        raise NoParametersFoundException

    def get_rabi_params(self, qcid):
        """
        Gets the current Rabi experiment parameters for a specific qubit on a specific device.

        :param int qcid: Unique identifier of the qubit in question
        :return: A dictionary with the parameter info
        :rtype: dict
        """
        def rabi_parse(qc):
            rabi_params = qc['rabi_params']
            return {
                'start': rabi_params['start'],
                'stop': rabi_params['stop'],
                'step': rabi_params['step'],
                'time': rabi_params['time'],
            }
        return self.get_params(qcid, rabi_parse)

    def get_ramsey_params(self, qcid):
        """
        Gets the current Ramsey experiment parameters for a specific qubit on a specific device.

        :param int qcid: Unique identifier of the qubit in question
        :return: A dictionary with the parameter info
        :rtype: dict
        """
        def ramsey_parse(qc):
            ramsey_params = qc['ramsey_params']
            return {
                'start': ramsey_params['start'],
                'stop': ramsey_params['stop'],
                'step': ramsey_params['step'],
                'detuning': ramsey_params['detuning'],
            }
        return self.get_params(qcid, ramsey_parse)

    def get_t1_params(self, qcid):
        """
        Gets the current T1 experiment parameters for a specific qubit on a specific device.

        :param int qcid: Unique identifier of the qubit in question
        :return: A dictionary with the parameter info
        :rtype: dict
        """
        def t1_parse(qc):
            t1_params = qc['t1_params']
            return {
                'start': t1_params['start'],
                'stop': t1_params['stop'],
                'num_pts': t1_params['num_pts'],
            }
        return self.get_params(qcid, t1_parse)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for working with Pauli algebras.
"""

from __future__ import division
from itertools import product
import numpy as np
import copy
from .quil import Program
from .gates import H, RZ, RX, CNOT, X, PHASE
from . import quilbase as pqb
from numbers import Number
from collections import Sequence
import warnings
from six import integer_types
from six.moves import range

PAULI_OPS = ["X", "Y", "Z", "I"]
PAULI_PROD = {'ZZ': 'I', 'YY': 'I', 'XX': 'I', 'II': 'I',
              'XY': 'Z', 'XZ': 'Y', 'YX': 'Z', 'YZ': 'X', 'ZX': 'Y',
              'ZY': 'X', 'IX': 'X', 'IY': 'Y', 'IZ': 'Z',
              'ZI': 'Z', 'YI': 'Y', 'XI': 'X',
              'X': 'X', 'Y': 'Y', 'Z': 'Z', 'I': 'I'}
PAULI_COEFF = {'ZZ': 1.0, 'YY': 1.0, 'XX': 1.0, 'II': 1.0,
               'XY': 1.0j, 'XZ': -1.0j, 'YX': -1.0j, 'YZ': 1.0j, 'ZX': 1.0j,
               'ZY': -1.0j, 'IX': 1.0, 'IY': 1.0, 'IZ': 1.0, 'ZI': 1.0,
               'YI': 1.0, 'XI': 1.0,
               'X': 1.0, 'Y': 1.0, 'Z': 1.0, 'I': 1.0}


class UnequalLengthWarning(Warning):
    def __init__(self, *args, **kwargs):
        super(UnequalLengthWarning, self).__init__(*args, **kwargs)


class PauliTerm(object):
    """A term is a product of Pauli operators operating on different qubits.
    """

    def __init__(self, op, index, coefficient=1.0):
        """ Create a new Pauli Term with a Pauli operator at a particular index and a leading
        coefficient.

        :param string op: The Pauli operator as a string "X", "Y", "Z", or "I"
        :param int index: The qubit index that that operator is applied to.
        :param float coefficient: The coefficient multiplying the operator, e.g. 1.5 * Z_1
        """
        assert op in PAULI_OPS
        assert isinstance(index, integer_types) and index >= 0

        self._ops = {}
        if op != "I":
            self._ops[index] = op
        if not isinstance(coefficient, Number):
            raise ValueError("coefficient of PauliTerm must be a Number.")
        self.coefficient = complex(coefficient)
        self._id = None

    def id(self):
        """
        Returns the unique identifier string for the PauliTerm (ignoring the coefficient).
         Used in the simplify method of PauliSum.

        :return: The unique identifier for this term.
        :rtype: string
        """
        if self._id is not None:
            return self._id
        else:
            s = ""
            for index in sorted(self._ops.keys()):
                s += "%s%s" % (self[index], index)
            self._id = s
            return s

    def __eq__(self, other):
        if not isinstance(other, (PauliTerm, PauliSum)):
            raise TypeError("Can't compare PauliTerm with object of type {}.".format(type(other)))
        elif isinstance(other, PauliSum):
            return other == self
        else:
            return self.id() == other.id() and np.isclose(self.coefficient, other.coefficient)

    def __ne__(self, other):
        # x!=y and x<>y call __ne__() instead of negating __eq__
        # This is only a weirdness in python2 as in python 3 __ne__ defaults to the inversion of
        # __eq__
        return not self.__eq__(other)

    def __len__(self):
        """
        The length of the PauliTerm is the number of Pauli operators in the term. A term that
        consists of only a scalar has a length of zero.
        """
        return len(self._ops)

    def copy(self):
        """
        Properly creates a new PauliTerm, with a completely new dictionary
        of operators
        """
        new_term = PauliTerm("I", 0, 1.0)  # create new object
        # manually copy all attributes over
        for key in self.__dict__.keys():
            val = self.__dict__[key]
            if isinstance(val, (dict, list, set)):  # mutable types
                new_term.__dict__[key] = copy.copy(val)
            else:  # immutable types
                new_term.__dict__[key] = val

        return new_term

    def get_qubits(self):
        """Gets all the qubits that this PauliTerm operates on.
        """
        # sort the keys to get a deterministic iteration order over qubits
        return sorted(self._ops.keys())

    def __getitem__(self, i):
        return self._ops.get(i, "I")

    def __iter__(self):
        for i in self.get_qubits():
            yield i, self[i]

    def _multiply_factor(self, factor, index):
        new_term = PauliTerm("I", 0)
        new_coeff = self.coefficient
        new_ops = self._ops.copy()

        ops = self[index] + factor
        new_op = PAULI_PROD[ops]
        if new_op != "I":
            new_ops[index] = new_op
        else:
            del new_ops[index]
        new_coeff *= PAULI_COEFF[ops]

        new_term._ops = new_ops
        new_term.coefficient = new_coeff

        return new_term

    def __mul__(self, term):
        """Multiplies this Pauli Term with another PauliTerm, PauliSum, or number according to the
        Pauli algebra rules.

        :param term: (PauliTerm or PauliSum or Number) A term to multiply by.
        :returns: The product of this PauliTerm and term.
        :rtype: PauliTerm
        """
        if isinstance(term, Number):
            return term_with_coeff(self, self.coefficient * term)
        elif isinstance(term, PauliSum):
            return (PauliSum([self]) * term).simplify()
        else:
            new_term = PauliTerm("I", 0, 1.0)
            new_term._ops = self._ops.copy()
            new_coeff = self.coefficient * term.coefficient
            for index, op in term:
                new_term = new_term._multiply_factor(op, index)

            return term_with_coeff(new_term, new_term.coefficient * new_coeff)

    def __rmul__(self, other):
        """Multiplies this PauliTerm with another object, probably a number.

        :param other: A number or PauliTerm to multiply by
        :returns: A new PauliTerm
        :rtype: PauliTerm
        """
        assert isinstance(other, Number)
        return self * other

    def __pow__(self, power):
        """Raises this PauliTerm to power.

        :param int power: The power to raise this PauliTerm to.
        :return: The power-fold product of power.
        :rtype: PauliTerm
        """
        if not isinstance(power, int) or power < 0:
            raise ValueError("The power must be a non-negative integer.")
        result = 1

        identities = [PauliTerm('I', qubit) for qubit in self.get_qubits()]
        if not identities:
            # There weren't any nontrivial operators
            return term_with_coeff(self, 1)
        for identity in identities:
            result *= identity
        for _ in range(power):
            result *= self
        return result

    def __add__(self, other):
        """Adds this PauliTerm with another one.

        :param other: A PauliTerm object or a Number
        :returns: A PauliSum object representing the sum of this PauliTerm and other
        :rtype: PauliSum
        """
        if isinstance(other, Number):
            return self + PauliTerm("I", 0, other)
        elif isinstance(other, PauliSum):
            return other + self
        else:
            new_sum = PauliSum([self, other])
            return new_sum.simplify()

    def __radd__(self, other):
        """Adds this PauliTerm with a Number.

        :param other: A PauliTerm object or a Number
        :returns: A new PauliTerm
        :rtype: PauliTerm
        """
        assert isinstance(other, Number)
        return self + other

    def __sub__(self, other):
        """Subtracts a PauliTerm from this one.

        :param other: A PauliTerm object or a Number
        :returns: A PauliSum object representing the difference of this PauliTerm and term
        :rtype: PauliSum
        """
        return self + -1. * other

    def __rsub__(self, other):
        """Subtracts this PauliTerm from a Number or PauliTerm.

        :param other: A PauliTerm object or a Number
        :returns: A PauliSum object representing the difference of this PauliTerm and term
        :rtype: PauliSum
        """
        return other + -1. * self

    def __str__(self):
        term_strs = []
        for index in sorted(self._ops.keys()):
            term_strs.append("%s%s" % (self[index], index))

        if len(term_strs) == 0:
            term_strs.append("I")
        out = "%s*%s" % (self.coefficient, '*'.join(term_strs))
        return out

    @classmethod
    def from_list(cls, terms_list, coefficient=1.0):
        """
        Allocates a Pauli Term from a list of operators and indices. This is more efficient than
        multiplying together individual terms.

        :param list terms_list: A list of tuples, e.g. [("X", 0), ("Y", 1)]
        :return: PauliTerm
        """
        pterm = PauliTerm("I", 0)
        assert all([op[0] in PAULI_OPS for op in terms_list])

        indices = [op[1] for op in terms_list]
        assert all([isinstance(index, integer_types) and index >= 0 for index in indices])

        # this is because from_list doesn't call simplify in order to be more efficient.
        if len(set(indices)) != len(indices):
            raise ValueError("Elements of PauliTerm that are allocated using from_list must "
                             "be on disjoint qubits. Use PauliTerm multiplication to simplify "
                             "terms instead.")

        for op, index in terms_list:
            if op != "I":
                pterm._ops[index] = op
        if not isinstance(coefficient, Number):
            raise ValueError("coefficient of PauliTerm must be a Number.")
        pterm.coefficient = complex(coefficient)
        return pterm


# For convenience, a shorthand for several operators.
ID = PauliTerm("I", 0, 1)
"""
The identity Pauli Term.
"""
ZERO = PauliTerm("I", 0, 0)
"""
The zero Pauli Term.
"""


def sI(q):
    """
    A function that returns the identity operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("I", q)


def sX(q):
    """
    A function that returns the sigma_X operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("X", q)


def sY(q):
    """
    A function that returns the sigma_Y operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("Y", q)


def sZ(q):
    """
    A function that returns the sigma_Z operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("Z", q)


def term_with_coeff(term, coeff):
    """
    Change the coefficient of a PauliTerm.

    :param PauliTerm term: A PauliTerm object
    :param Number coeff: The coefficient to set on the PauliTerm
    :returns: A new PauliTerm that duplicates term but sets coeff
    :rtype: PauliTerm
    """
    if not isinstance(coeff, Number):
        raise ValueError("coeff must be a Number")
    new_pauli = term.copy()
    # We cast to a complex number to ensure that internally the coefficients remain compatible.
    new_pauli.coefficient = complex(coeff)
    return new_pauli


class PauliSum(object):
    """A sum of one or more PauliTerms.
    """

    def __init__(self, terms):
        """
        :param Sequence terms: A Sequence of PauliTerms.
        """
        if not (isinstance(terms, Sequence) and
                all([isinstance(term, PauliTerm) for term in terms])):
            raise ValueError("PauliSum's are currently constructed from Sequences of PauliTerms.")
        if len(terms) == 0:
            self.terms = [0.0 * ID]
        else:
            self.terms = terms

    def __eq__(self, other):
        """Equality testing to see if two PauliSum's are equivalent.

        :param PauliSum other: The PauliSum to compare this PauliSum with.
        :return: True if other is equivalent to this PauliSum, False otherwise.
        :rtype: bool
        """
        if not isinstance(other, (PauliTerm, PauliSum)):
            raise TypeError("Can't compare PauliSum with object of type {}.".format(type(other)))
        elif isinstance(other, PauliTerm):
            return self == PauliSum([other])
        elif len(self.terms) != len(other.terms):
            warnings.warn(UnequalLengthWarning("These PauliSums have a different number of terms."))
            return False
        return all([term == other.terms[i] for i, term in enumerate(self.terms)])

    def __ne__(self, other):
        """Inequality testing to see if two PauliSum's are not equivalent.

        :param PauliSum other: The PauliSum to compare this PauliSum with.
        :return: False if other is equivalent to this PauliSum, True otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self):
        return " + ".join([str(term) for term in self.terms])

    def __len__(self):
        """
        The length of the PauliSum is the number of PauliTerms in the sum.
        """
        return len(self.terms)

    def __getitem__(self, item):
        """
        :param int item: The index of the term in the sum to return
        :return: The PauliTerm at the index-th position in the PauliSum
        :rtype: PauliTerm
        """
        return self.terms[item]

    def __iter__(self):
        return self.terms.__iter__()

    def __mul__(self, other):
        """
        Multiplies together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the multiplication.
        :rtype: PauliSum
        """
        if not isinstance(other, (Number, PauliTerm, PauliSum)):
            raise ValueError("Cannot multiply PauliSum by term that is not a Number, PauliTerm, or"
                             "PauliSum")
        elif isinstance(other, PauliSum):
            other_terms = other.terms
        else:
            other_terms = [other]
        new_terms = [lterm * rterm for lterm, rterm in product(self.terms, other_terms)]
        new_sum = PauliSum(new_terms)
        return new_sum.simplify()

    def __rmul__(self, other):
        """
        Multiples together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the multiplication.
        :rtype: PauliSum
        """
        assert isinstance(other, Number)
        new_terms = [term.copy() for term in self.terms]
        for term in new_terms:
            term.coefficient *= other
        return PauliSum(new_terms).simplify()

    def __pow__(self, power):
        """Raises this PauliSum to power.

        :param int power: The power to raise this PauliSum to.
        :return: The power-th power of this PauliSum.
        :rtype: PauliSum
        """
        if not isinstance(power, int) or power < 0:
            raise ValueError("The power must be a non-negative integer.")
        result = 1

        if not self.get_qubits():
            # There aren't any nontrivial operators
            terms = [term_with_coeff(term, 1) for term in self.terms]
            for term in terms:
                result *= term
        else:
            for term in self.terms:
                for qubit_id in term.get_qubits():
                    result *= PauliTerm("I", qubit_id)

        for _ in range(power):
            result *= self
        return result

    def __add__(self, other):
        """
        Adds together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the addition.
        :rtype: PauliSum
        """
        if isinstance(other, PauliTerm):
            other = PauliSum([other])
        elif isinstance(other, Number):
            other = PauliSum([other * ID])
        new_terms = [term.copy() for term in self.terms]
        new_terms.extend(other.terms)
        new_sum = PauliSum(new_terms)
        return new_sum.simplify()

    def __radd__(self, other):
        """
        Adds together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the addition.
        :rtype: PauliSum
        """
        assert isinstance(other, Number)
        return self + other

    def __sub__(self, other):
        """
        Finds the difference of this PauliSum with PauliSum, PauliTerm or Number objects. The new
        term is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the subtraction.
        :rtype: PauliSum
        """
        return self + -1. * other

    def __rsub__(self, other):
        """
        Finds the different of this PauliSum with PauliSum, PauliTerm or Number objects. The new
        term is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the subtraction.
        :rtype: PauliSum
        """
        return other + -1. * self

    def get_qubits(self):
        """
        The support of all the operators in the PauliSum object.

        :returns: A list of all the qubits in the sum of terms.
        :rtype: list
        """
        return list(set().union(*[term.get_qubits() for term in self.terms]))

    def simplify(self):
        """
        Simplifies the sum of Pauli operators according to Pauli algebra rules.
        """
        def coalesce(d):
            terms = []
            for k in sorted(d):
                term_list = d[k]
                if (len(term_list) == 1 and not
                        np.isclose(term_list[0].coefficient, 0.0)):
                    terms.append(term_list[0])
                else:
                    coeff = sum(t.coefficient for t in term_list)
                    if not np.isclose(coeff, 0.0):
                        terms.append(term_with_coeff(term_list[0], coeff))
            return PauliSum(terms)

        like_terms = {}
        for term in self.terms:
            id = term.id()
            if id not in like_terms:
                like_terms[id] = [term]
            else:
                like_terms[id] = like_terms[id] + [term]

        return coalesce(like_terms)


def check_commutation(pauli_list, pauli_two):
    """
    Check if commuting a PauliTerm commutes with a list of other terms by natural calculation.
    Derivation similar to arXiv:1405.5749v2 fo the check_commutation step in
    the Raesi, Wiebe, Sanders algorithm (arXiv:1108.4318, 2011).

    :param list pauli_list: A list of PauliTerm objects
    :param PauliTerm pauli_two_term: A PauliTerm object
    :returns: True if pauli_two object commutes with pauli_list, False otherwise
    :rtype: bool
    """
    def coincident_parity(p1, p2):
        non_similar = 0
        p1_indices = set(p1._ops.keys())
        p2_indices = set(p2._ops.keys())
        for idx in p1_indices.intersection(p2_indices):
            if p1[idx] != p2[idx]:
                non_similar += 1
        return non_similar % 2 == 0

    for i, term in enumerate(pauli_list):
        if not coincident_parity(term, pauli_two):
            return False
    return True


def commuting_sets(pauli_terms, nqubits):
    """Gather the Pauli terms of pauli_terms variable into commuting sets

    Uses algorithm defined in (Raeisi, Wiebe, Sanders, arXiv:1108.4318, 2011)
    to find commuting sets. Except uses commutation check from arXiv:1405.5749v2

    :param PauliSum pauli_terms: A PauliSum object
    :returns: List of lists where each list contains a commuting set
    :rtype: list
    """

    m_terms = len(pauli_terms.terms)
    m_s = 1
    groups = []
    groups.append([pauli_terms.terms[0]])
    for j in range(1, m_terms):
        isAssigned_bool = False
        for p in range(m_s):  # check if it commutes with each group
            if isAssigned_bool is False:

                if check_commutation(groups[p], pauli_terms.terms[j]):
                    isAssigned_bool = True
                    groups[p].append(pauli_terms.terms[j])
        if isAssigned_bool is False:
            m_s += 1
            groups.append([pauli_terms.terms[j]])
    return groups


def is_identity(term):
    """
    Check if Pauli Term is a scalar multiple of identity

    :param PauliTerm term: A PauliTerm object
    :returns: True if the PauliTerm is a scalar multiple of identity, false otherwise
    :rtype: bool
    """
    return len(term) == 0


def exponentiate(term):
    """
    Creates a pyQuil program that simulates the unitary evolution exp(-1j * term)

    :param PauliTerm term: Tests is a PauliTerm is the identity operator
    :returns: A Program object
    :rtype: Program
    """
    return exponential_map(term)(1.0)


def exponential_map(term):
    """
    Creates map alpha -> exp(-1j*alpha*term) represented as a Program.

    :param PauliTerm term: Tests is a PauliTerm is the identity operator
    :returns: Program
    :rtype: Program
    """
    if not np.isclose(np.imag(term.coefficient), 0.0):
        raise TypeError("PauliTerm coefficient must be real")

    coeff = term.coefficient.real
    term.coefficient = term.coefficient.real

    def exp_wrap(param):
        prog = Program()
        if is_identity(term):
            prog.inst(X(0))
            prog.inst(PHASE(-param * coeff)(0))
            prog.inst(X(0))
            prog.inst(PHASE(-param * coeff)(0))
        else:
            prog += _exponentiate_general_case(term, param)
        return prog

    return exp_wrap


def _exponentiate_general_case(pauli_term, param):
    """
    Returns a Quil (Program()) object corresponding to the exponential of
    the pauli_term object, i.e. exp[-1.0j * param * pauli_term]

    :param PauliTerm pauli_term: A PauliTerm to exponentiate
    :param float param: scalar, non-complex, value
    :returns: A Quil program object
    :rtype: Program
    """
    def reverse_hack(p):
        # A hack to produce a *temporary* program which reverses p.
        def translate(tup):
            action, obj = tup
            if tup == pqb.ACTION_RELEASE_QUBIT:
                return (pqb.ACTION_INSTANTIATE_QUBIT, obj)
            elif tup == pqb.ACTION_INSTANTIATE_QUBIT:
                return (pqb.ACTION_RELEASE_QUBIT, obj)
            else:
                return tup
        revp = Program()
        revp.actions = [translate(action) for action in reversed(p.actions)]
        return revp

    quil_prog = Program()
    change_to_z_basis = Program()
    change_to_original_basis = Program()
    cnot_seq = Program()
    prev_index = None
    highest_target_index = None

    for index, op in pauli_term:
        if 'X' == op:
            change_to_z_basis.inst(H(index))
            change_to_original_basis.inst(H(index))

        elif 'Y' == op:
            change_to_z_basis.inst(RX(np.pi / 2.0)(index))
            change_to_original_basis.inst(RX(-np.pi / 2.0)(index))

        elif 'I' == op:
            continue

        if prev_index is not None:
            cnot_seq.inst(CNOT(prev_index, index))

        prev_index = index
        highest_target_index = index

    # building rotation circuit
    quil_prog += change_to_z_basis
    quil_prog += cnot_seq
    quil_prog.inst(RZ(2.0 * pauli_term.coefficient * param)(highest_target_index))
    quil_prog += reverse_hack(cnot_seq)
    quil_prog += change_to_original_basis

    return quil_prog


def suzuki_trotter(trotter_order, trotter_steps):
    """
    Generate trotterization coefficients for a given number of Trotter steps.

    U = exp(A + B) is approximated as exp(w1*o1)exp(w2*o2)... This method returns
    a list [(w1, o1), (w2, o2), ... , (wm, om)] of tuples where o=0 corresponds
    to the A operator, o=1 corresponds to the B operator, and w is the
    coefficient in the exponential. For example, a second order Suzuki-Trotter
    approximation to exp(A + B) results in the following
    [(0.5/trotter_steps, 0), (1/trotteri_steps, 1),
    (0.5/trotter_steps, 0)] * trotter_steps.

    :param int trotter_order: order of Suzuki-Trotter approximation
    :param int trotter_steps: number of steps in the approximation
    :returns: List of tuples corresponding to the coefficient and operator
              type: o=0 is A and o=1 is B.
    :rtype: list
    """
    p1 = p2 = p4 = p5 = 1.0 / (4 - (4**(1. / 3)))
    p3 = 1 - 4 * p1
    trotter_dict = {1: [(1, 0), (1, 1)],
                    2: [(0.5, 0), (1, 1), (0.5, 0)],
                    3: [(7.0 / 24, 0), (2.0 / 3.0, 1), (3.0 / 4.0, 0), (-2.0 / 3.0, 1),
                        (-1.0 / 24, 0), (1.0, 1)],
                    4: [(p5 / 2, 0), (p5, 1), (p5 / 2, 0),
                        (p4 / 2, 0), (p4, 1), (p4 / 2, 0),
                        (p3 / 2, 0), (p3, 1), (p3 / 2, 0),
                        (p2 / 2, 0), (p2, 1), (p2 / 2, 0),
                        (p1 / 2, 0), (p1, 1), (p1 / 2, 0)]}

    order_slices = [(x0 / trotter_steps, x1) for x0, x1 in trotter_dict[trotter_order]]
    order_slices = order_slices * trotter_steps
    return order_slices


def is_zero(pauli_object):
    """
    Tests to see if a PauliTerm or PauliSum is zero.

    :param pauli_object: Either a PauliTerm or PauliSum
    :returns: True if PauliTerm is zero, False otherwise
    :rtype: bool
    """
    if isinstance(pauli_object, PauliTerm):
        if pauli_object.id() == '':
            return True
        else:
            return False
    elif isinstance(pauli_object, PauliSum):
        if len(pauli_object.terms) == 1 and pauli_object.terms[0].id() == '':
            return True
        else:
            return False
    else:
        raise TypeError("is_zero only checks PauliTerms and PauliSum objects!")


def trotterize(first_pauli_term, second_pauli_term, trotter_order=1,
               trotter_steps=1):
    """
    Create a Quil program that approximates exp( (A + B)t) where A and B are
    PauliTerm operators.

    :param PauliTerm first_pauli_term: PauliTerm denoted `A`
    :param PauliTerm second_pauli_term: PauliTerm denoted `B`
    :param int trotter_order: Optional argument indicating the Suzuki-Trotter
                          approximation order--only accepts orders 1, 2, 3, 4.
    :param int trotter_steps: Optional argument indicating the number of products
                          to decompose the exponential into.

    :return: Quil program
    :rtype: Program
    """

    if not (1 <= trotter_order < 5):
        raise ValueError("trotterize only accepts trotter_order in {1, 2, 3, 4}.")

    commutator = (first_pauli_term * second_pauli_term) +\
                 (-1 * second_pauli_term * first_pauli_term)

    prog = Program()
    if is_zero(commutator):
        param_exp_prog_one = exponential_map(first_pauli_term)
        exp_prog = param_exp_prog_one(1)
        prog += exp_prog
        param_exp_prog_two = exponential_map(second_pauli_term)
        exp_prog = param_exp_prog_two(1)
        prog += exp_prog
        return prog

    order_slices = suzuki_trotter(trotter_order, trotter_steps)
    for coeff, operator in order_slices:
        if operator == 0:
            param_prog = exponential_map(coeff * first_pauli_term)
            exp_prog = param_prog(1)
            prog += exp_prog
        else:
            param_prog = exponential_map(coeff * second_pauli_term)
            exp_prog = param_prog(1)
            prog += exp_prog
    return prog



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Contains objects and functions to manage allocation and de-allocation of quantum and classical
resources in pyQuil.
"""

from itertools import count
from .quil_atom import QuilAtom
from copy import copy
from six import integer_types


class AbstractQubit(QuilAtom):
    """
    Representation of a qubit.
    """

    def index(self):
        raise NotImplementedError()


class DirectQubit(AbstractQubit):
    """
    A reference to a particularly indexed qubit.
    """

    def __init__(self, index):
        if not isinstance(index, integer_types) or index < 0:
            raise TypeError("index should be a non-negative int")
        self._index = index

    def index(self):
        return self._index

    def __repr__(self):
        return "<DirectQubit %d>" % self.index()

    def __str__(self):
        return str(self.index())

    def __eq__(self, other):
        if not isinstance(other, DirectQubit):
            raise TypeError("Can only compare DirectQubit instances with DirectQubit instances")
        return self.index() == other.index()

    def __ne__(self, other):
        # x!=y and x<>y call __ne__() instead of negating __eq__
        # This is only a weirdness in python2 as in python 3 __ne__ defaults to the inversion of
        # __eq__
        return not self.__eq__(other)


class Qubit(AbstractQubit):
    """
    A qubit whose index is determined by a ResourceManager.
    """

    def __init__(self, resource_manager):
        self.resource_manager = resource_manager
        self.live = True
        self.assignment = None

    def index(self):
        """
        :returns: The index of this qubit.
        :rtype: int
        """
        if not instantiated(self):
            raise RuntimeError("Can't get the index of an unassigned qubit.")
        return self.assignment

    def __repr__(self):
        return "<Qubit %X>" % id(self)

    def __str__(self):
        return str(self.index())


def check_live_qubit(qubit):
    """
    Ensure that an object is a live qubit. Raises errors if not.

    :param qubit: A qubit, either a DirectQubit or Qubit object.
    """
    if isinstance(qubit, DirectQubit):
        pass
    elif isinstance(qubit, Qubit):
        if not qubit.live:
            raise RuntimeError("qubit is not live when it is expected to be")
    else:
        raise TypeError("qubit should be an AbstractQubit instance")


def instantiated(qubit):
    """
    Is the qubit instantiated?

    :param AbstractQubit qubit: A AbstractQubit instance.
    :return: A boolean.
    :rtype: bool
    """
    return qubit.assignment is not None


class ResourceManager(object):
    def __init__(self):
        self.live_qubits = []
        self.dead_qubits = []
        self.in_use = {}

    def reset(self):
        """
        Frees all qubits, resetting this ResourceManager.
        """
        # A copy is necessary here so that the modifications to the list don't affect the traversal.
        for qubit in copy(self.live_qubits):
            self.free_qubit(qubit)
            qubit.resource_manager = None
        self.live_qubits = []
        self.dead_qubits = []
        self.in_use = {}

    def allocate_qubit(self):
        """
        Create a new Qubit object.

        :return: A Qubit instance.
        :rtype: Qubit
        """
        new_q = Qubit(self)
        self.live_qubits.append(new_q)
        return new_q

    def free_qubit(self, q):
        """
        Free the previously allocated qubit.

        :param Qubit q: A live Qubit that was previously allocated.
        """
        if q not in self.live_qubits:
            raise RuntimeError("Qubit {} is not part of the ResourceManager".format(q))
        self.live_qubits.remove(q)
        q.live = False
        self.dead_qubits.append(q)

    def instantiate(self, qubit):
        """
        Assign a number to a qubit.

        :param Qubit qubit: A Qubit object.
        """

        def find_available_qubit(d):
            # Just do a linear search.
            for i in count(start=0, step=1):
                if not d.get(i, False):
                    return i

        if not (isinstance(qubit, Qubit)):
            raise TypeError(qubit, "should be a Qubit")

        qubit.assignment = find_available_qubit(self.in_use)
        self.in_use[qubit.assignment] = qubit

    def uninstantiate_index(self, i):
        """
        Free up an index that was previously instantiated.

        :param int i: The index.
        """
        q = self.in_use.get(i, False)
        if q and isinstance(q, Qubit):
            del self.in_use[i]


def merge_resource_managers(rm1, rm2):
    """
    Merge two resource managers into a new resource manager. All qubits in the old managers will
    point to the new one. The in_use labels of the second resource manager have priority.

    :param ResourceManager rm1: A ResourceManager.
    :param ResourceManager rm2: A ResourceManager.
    :return: A merged ResourceManager.
    :rtype: ResourceManager
    """
    rm = ResourceManager()
    rm.live_qubits = rm1.live_qubits + rm2.live_qubits
    rm.dead_qubits = rm1.dead_qubits + rm2.dead_qubits
    for q in rm.dead_qubits + rm.live_qubits:
        q.resource_manager = rm
    rm.in_use.update(rm1.in_use)
    rm.in_use.update(rm2.in_use)
    return rm



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for reading configuration information about api keys and user ids.
"""

from __future__ import print_function
from six.moves.configparser import ConfigParser, NoSectionError, NoOptionError
from os.path import expanduser

from os import getenv

import sys


class PyquilConfig(object):
    DEFAULT_PYQUIL_CONFIG_PATH = expanduser('~/.pyquil_config')
    PYQUIL_CONFIG_PATH = getenv('PYQUIL_CONFIG', DEFAULT_PYQUIL_CONFIG_PATH)

    SECTION = "Rigetti Forest"
    ENDPOINT = "url"
    API_KEY = "key"
    USER_ID = "user_id"

    DEFAULT_ENDPOINT = "https://api.rigetti.com/qvm"

    def __init__(self):
        self.configparser = ConfigParser()

        if len(self.configparser.read(self.PYQUIL_CONFIG_PATH)) == 0:
            print("! WARNING:\n"
                  "!   There was an issue finding your pyQuil config file.\n"
                  "!   Have you run the pyquil-config-setup command yet?\n"
                  "! See the getting started guide at https://go.rigetti.com/getting-started",
                  file=sys.stderr)

    @property
    def endpoint(self):
        endpoint = self._env_or_config('QVM_URL', self.ENDPOINT)
        return endpoint if endpoint else self.DEFAULT_ENDPOINT

    @property
    def api_key(self):
        return self._env_or_config('QVM_API_KEY', self.API_KEY)

    @property
    def user_id(self):
        return self._env_or_config('QVM_USER_ID', self.USER_ID)

    def _env_or_config(self, env, name):
        """
        Get the value of the environment variable or config file value.
        The environment variable takes precedence.

        :param env: The environment variable name.
        :param name: The config file key.
        :return: The value or None if not found
        """
        env_val = getenv(env)
        if env_val is not None:
            return env_val

        try:
            return self.configparser.get(self.SECTION, name)
        except (NoSectionError, NoOptionError, KeyError):
            return None



#!/usr/bin/env python
##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from setuptools import setup
from pyquil import __version__

setup(
    name="pyquil",
    version=__version__,
    author="Rigetti Computing",
    author_email="softapps@rigetti.com",
    description="A Python library to generate Quantum Instruction Language (Quil) Programs.",
    url="https://github.com/rigetticomputing/pyquil.git",
    packages=['pyquil', 'pyquil.setup'],
    license="LICENSE",
    install_requires=[
        'requests >= 2.4.2',
        'numpy >= 1.10',
        'matplotlib >= 1.5',
        'urllib3 >= 1.21.1',
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest >= 3.0.0',
        'mock',
    ],
    test_suite='pyquil.tests',
    entry_points={
        'console_scripts': ['pyquil-config-setup=pyquil.setup.pyquil_config_setup:main']
    },
    keywords='quantum quil programming hybrid'
)



import matplotlib.pyplot as plt
import numpy as np
from lmfit import Model
from lmfit.models import update_param_vals

from job_results import RabiResult, RamseyResult, T1Result


def analog_plot(job_result):
    """
    Make a plot, using matplotlib, of an experimental result.
    :param JobResult job_result:
    :return: None
    """
    xs, amps = job_result.decode()
    if isinstance(job_result, RabiResult):
        plt.title("Qubit Rabi")
        plt.ylabel("Response")
        plt.xlabel("Pulse Amplitude (mV)")
    elif isinstance(job_result, RamseyResult):
        plt.title("Qubit Ramsey")
        plt.ylabel("Response")
        plt.xlabel("Pulse Delay (us)")
        xs = np.asarray(xs) * 1.e6
    elif isinstance(job_result, T1Result):
        plt.title("Qubit T1")
        plt.ylabel("Response")
        plt.xlabel("Readout Delay (us)")
        xs = np.asarray(xs) * 1.e6
    plt.plot(xs, amps, 'bo')
    plt.show()


##################################################################
# T1 decay
##################################################################
def fn_T1_decay(x, baseline, amplitude, T1):
    return baseline + amplitude * np.exp(-x / T1)


MIN_DATA_POINTS = 5


class T1DecayModel(Model):
    __doc__ = """Class for T1 decay model"""

    def __init__(self, *args, **kwargs):
        super(T1DecayModel, self).__init__(fn_T1_decay, *args, **kwargs)

    __model__ = Model(fn_T1_decay)

    def guess(self, x, y, **kwargs):
        """Takes x and y data and generates initial guesses for
        fit parameters.
        """
        if y[0] > y[-1]:
            b_guess = y.min()
            a_guess = 1.0 * (y.max() - b_guess)
        else:
            b_guess = y.max()
            a_guess = 1.0 * (y.min() - b_guess)
        y2 = y - b_guess
        x2 = x - x[0]
        sy2 = sum(y2)
        if sy2 == 0.:
            # prevent division by 0
            t_guess = 100 * 1.e-6
        else:
            t_guess = 0.5 * np.dot(x2, y2) / sy2  # 1st moment of x: sum(x*f(x)) / sum(f(x))

        pars = self.make_params(baseline=b_guess, amplitude=a_guess, T1=t_guess)
        pars['amplitude'].value = kwargs.pop('amplitude_guess', a_guess)
        pars['baseline'].value = kwargs.pop('baseline_guess', b_guess)
        pars['T1'].value = kwargs.pop('T1_guess', t_guess)
        return update_param_vals(pars, self.prefix, **kwargs)

    def do_fit(self, x, y, errs=None, **kwargs):
        """Performs a fit.
        """
        par = self.guess(x, y, **kwargs)
        if errs is not None:
            fit = self.fit(y, x=x, weights=1.0 / errs, params=par)
        else:
            fit = self.fit(y, x=x, params=par)
        return fit

    def report_fit(self, x, y, **kwargs):
        """Reports the results of a fit. May change depending on what we
        want this to return.
        """

        if not len(x) == len(y):
            raise ValueError("Lengths of x and y arrays must be equal.")
        if not len(x) > MIN_DATA_POINTS:
            raise ValueError("You must provide more than {} data points.".format(MIN_DATA_POINTS))

        fit = self.do_fit(x, y, **kwargs)
        t_fit = fit.params['T1']
        return fit, (t_fit.value, t_fit.stderr)

##################################################################
# T2 Ramsey
##################################################################


def fn_T2_Ramsey(x, baseline, amplitude, T2, detuning, x0=0.0):
    """Ramsey lineshape.
    """
    return baseline + amplitude * np.exp(-x / T2) * np.cos(2 * np.pi * detuning * (x - x0))


class T2RamseyModel(Model):
    __doc__ = """Class for T2 Ramsey model"""

    def __init__(self, *args, **kwargs):
        super(T2RamseyModel, self).__init__(fn_T2_Ramsey, *args, **kwargs)

    __model__ = Model(fn_T2_Ramsey)

    def guess(self, x, y, **kwargs):
        """Takes x and y data and generates initial guesses for
        fit parameters.
        """
        ym = y.min()
        yM = y.max()
        a_guess = 0.5 * (yM - ym) * 1.2
        b_guess = 0.5 * (yM + ym)
        t_guess = 0.25 * x[-1]
        d_guess = get_freq_from_fft(x, y, b_guess)
        x_guess = 0.0

        pars = self.make_params(baseline=b_guess, amplitude=a_guess,
                                T2=t_guess, detuning=d_guess, x0=x_guess)
        pars['amplitude'].value = kwargs.pop('amplitude_guess', a_guess)
        pars['baseline'].value = kwargs.pop('baseline_guess', b_guess)
        pars['T2'].value = kwargs.pop('T2_guess', t_guess)
        pars['detuning'].value = kwargs.pop('detuning_guess', d_guess)
        pars['x0'].value = kwargs.pop('x0_guess', x_guess)
        return update_param_vals(pars, self.prefix, **kwargs)

    def do_fit(self, x, y, errs=None, **kwargs):
        """Performs a fit.
        """
        par = self.guess(x, y, **kwargs)
        if errs is not None:
            fit = self.fit(y, x=x, weights=1.0 / errs, params=par, fit_kws={"nan_policy": "omit"})
        else:
            fit = self.fit(y, x=x, params=par, fit_kws={"nan_policy": "omit"})
        return fit

    def report_fit(self, x, y, **kwargs):
        """Reports the results of a fit. May change depending on what we
        want this to return.
        """

        if not len(x) == len(y):
            raise ValueError("Lengths of x and y arrays must be equal.")
        if not len(x) > MIN_DATA_POINTS:
            raise ValueError("You must provide more than {} data points.".format(MIN_DATA_POINTS))

        fit = self.do_fit(x, y, **kwargs)
        t_fit = fit.params['T2']
        d_fit = fit.params['detuning']
        return fit, (t_fit.value, t_fit.stderr), (d_fit.value, d_fit.stderr)


def get_freq_from_fft(x, y, offset=0.0):
    """Returns the frequency at which an fft is peaked in amplitude.
    Notes:
    - A constant offset is subtracted
    - The two largest frequency components are taken in case f=0 is still the
    largest components
    - A sine wave will have peaks at +/- the true frequency. This function
    returns the absolute value.
    """
    assert len(x) > 1
    f = np.fft.fftfreq(len(y), x[1] - x[0])
    A = np.abs(np.fft.fft(y - offset))
    f_max_2 = f[np.argpartition(A, -2)][-2:]  # f's for largest 2 elements of A
    if f_max_2[0] != 0.0:
        f_max = f_max_2[0]
    else:
        f_max = f_max_2[1]
    return np.abs(f_max)



import base64
import json
import struct
import time

from six import integer_types
import numpy as np

from pyquil.wavefunction import Wavefunction

OCTETS_PER_DOUBLE_FLOAT = 8
OCTETS_PER_COMPLEX_DOUBLE = 2 * OCTETS_PER_DOUBLE_FLOAT


def wait_for_job(res, ping_time=0.5):
    """
    Blocks execution and waits for an async Forest Job to complete.

    :param JobResult res: The JobResult object to wait for.
    :param ping_time: The interval (in seconds) at which to ping the server.
    :return: The completed JobResult
    """
    while not res.is_done():
        time.sleep(ping_time)
    return res


class JobResult(object):
    def __init__(self, qpu, success, result=None, payload=None):
        """
        :param QPUConnection qpu:
        :param bool success:
        :param dict result: JSON dictionary of the result message
        """
        self.qpu = qpu
        self.success = success
        self.result = result
        self.payload = payload

    def is_done(self):
        """
        Polls Forest API for status, then returns updated status.

        :return: Returns True if the Job is completed and False otherwise.
        :rtype: bool
        """
        self.get()
        return 'result' in self.result

    def job_id(self):
        """
        :return: Returns the id of this job
        :rtype: str
        """
        return self.result['jobId']

    def get(self):
        """
        Gets an update from the Forest API on the status of this job.
        :return: The JobResult object with fields updated.
        :rtype: JobResult
        """
        return self.qpu.get_job(self)

    def _update(self, success, result):
        self.success = success
        self.result = result
        return self

    def decode(self):
        """
        Decodes the result of the job.
        :return: Depends on the type of job. A JSON object.
        """
        return self.result['result']

    def __str__(self):
        return str(self.result)

    @classmethod
    def load_res(cls, qpu, response):
        result = response.json()
        return cls(qpu, response.ok, result)


class WavefunctionResult(JobResult):

    def decode(self):
        string_result = self.result['result']
        coef_string = base64.b64decode(string_result.encode("utf-8"))
        wvf, classical_mem = recover_complexes(coef_string=coef_string,
                                               classical_addresses=self.payload['addresses'])
        return wvf, classical_mem


def recover_complexes(coef_string, classical_addresses):
    """
    From a bit packed string, unpacks to get the wavefunction and classical measurement results
    :param coef_string:
    :param classical_addresses:
    :return:
    """
    num_octets = len(coef_string)
    num_addresses = len(classical_addresses)
    num_memory_octets = _round_to_next_multiple(num_addresses, 8) // 8
    num_wavefunction_octets = num_octets - num_memory_octets

    # Parse the classical memory
    mem = []
    for i in range(num_memory_octets):
        octet = struct.unpack('B', coef_string[i:i + 1])[0]
        mem.extend(_octet_bits(octet))

    mem = mem[0:num_addresses]

    # Parse the wavefunction
    wf = np.zeros(num_wavefunction_octets // OCTETS_PER_COMPLEX_DOUBLE, dtype=np.cfloat)
    for i, p in enumerate(range(num_memory_octets, num_octets, OCTETS_PER_COMPLEX_DOUBLE)):
        re_be = coef_string[p: p + OCTETS_PER_DOUBLE_FLOAT]
        im_be = coef_string[p + OCTETS_PER_DOUBLE_FLOAT: p + OCTETS_PER_COMPLEX_DOUBLE]
        re = struct.unpack('>d', re_be)[0]
        im = struct.unpack('>d', im_be)[0]
        wf[i] = complex(re, im)

    return Wavefunction(wf), mem


class RamseyResult(JobResult):
    pass


class RabiResult(JobResult):
    pass


class T1Result(JobResult):
    pass


def _round_to_next_multiple(n, m):
    """
    Round up the the next multiple.

    :param n: The number to round up.
    :param m: The multiple.
    :return: The rounded number
    """
    return n if n % m == 0 else n + m - n % m


def _octet_bits(o):
    """
    Get the bits of an octet.

    :param o: The octets.
    :return: The bits as a list in LSB-to-MSB order.
    :rtype: list
    """
    if not isinstance(o, integer_types):
        raise TypeError("o should be an int")
    if not (0 <= o <= 255):
        raise ValueError("o should be between 0 and 255 inclusive")
    bits = [0] * 8
    for i in range(8):
        if 1 == o & 1:
            bits[i] = 1
        o = o >> 1
    return bits
